/********************************************************************
*
* tgx library: Crazy clock demo.
*
* This example show how to use blitting of rotated/rescaled sprites
* with transparency. 
*
* EXAMPLE FOR TEENSY 4 / 4.1
*
* DISPLAY: ILI9341 (320x240)
*
********************************************************************/
//#define USE_ST7789
#define USE_ILI9488
// This example runs on teensy 4.0/4.1 with ILI9341 via SPI.
// the screen driver library : https://github.com/vindar/ILI9341_T4
//#include <ILI9341_T4.h>
#ifdef USE_ST7789
#include <ST7735_t3.h>      // Hardware-specific library
#include <ST7789_t3.h>      // Hardware-specific library
#define SPI_SPEED 40000000  // SPI speed
#define TFT_WIDTH 320
#define TFT_HEIGHT 240
#define IMAGE_CENTER ST7789_t3::CENTER
#elif defined(USE_ILI9488)
#include <ILI9488_t3.h>
#define TFT_WIDTH 480       //320
#define TFT_HEIGHT 320      //480
#define SPI_SPEED 20000000  // SPI speed
#define IMAGE_CENTER ILI9488_t3::CENTER
#else
#include <ILI9341_t3n.h>
#define IMAGE_CENTER ILI9341_t3n::CENTER
#define SPI_SPEED 40000000  // SPI speed
#define TFT_WIDTH 320       //240
#define TFT_HEIGHT 240      //320
#endif

// for now define TGX frame buffer to be size of the images defined...
// the tgx library
#include <tgx.h>
#include <font_tgx_OpenSans_Bold.h>


// let's not burden ourselves with the tgx:: prefix
using namespace tgx;

// the sprites
#include "antique_clock.h"  // the clock, RGB565 format, no transparency
#include "long_hand.h"      // long hand, RGB32 format (with transparency)
#include "small_hand.h"     // small hand, RGB32 format (with transparency)


//
// DEFAULT WIRING USING SPI 0 ON TEENSY 4/4.1
//
#define PIN_SCK 13   // mandatory
#define PIN_MISO 12  // mandatory
#define PIN_MOSI 11  // mandatory
#define PIN_DC 10    // mandatory, can be any pin but using pin 10 (or 36 or 37 on T4.1) provides greater performance

#define PIN_CS 9           // optional (but recommended), can be any pin.
#define PIN_RESET 6        // optional (but recommended), can be any pin.
#define PIN_BACKLIGHT 255  // optional, set this only if the screen LED pin is connected directly to the Teensy.
#define PIN_TOUCH_IRQ 255  // optional. set this only if the touchscreen is connected on the same SPI bus
#define PIN_TOUCH_CS 255   // optional. set this only if the touchscreen is connected on the same spi bus


//
// ALTERNATE WIRING USING SPI 1 ON TEENSY 4/4.1
//
//#define PIN_SCK     27      // mandatory
//#define PIN_MISO    1       // mandatory
//#define PIN_MOSI    26      // mandatory
//#define PIN_DC      0       // mandatory, can be any pin but using pin 0 (or 38 on T4.1) provides greater performance

//#define PIN_CS      30      // optional (but recommended), can be any pin.
//#define PIN_RESET   29      // optional (but recommended), can be any pin.
//#define PIN_BACKLIGHT 255   // optional, set this only if the screen LED pin is connected directly to the Teensy.
//#define PIN_TOUCH_IRQ 255   // optional. set this only if the touchscreen is connected on the same SPI bus
//#define PIN_TOUCH_CS  255   // optional. set this only if the touchscreen is connected on the same spi bus


//#define SPI_SPEED 40000000

#undef TFT_DC
#undef TFT_CS
#undef TFT_RST
#define TFT_DC 9
#define TFT_CS 32
#define TFT_RST 31
// used for XPT2046
#undef TOUCH_CS
#undef TOUCH_TIRQ
#define TOUCH_CS 26
#define TOUCH_TIRQ 27


// the screen driver object
//ILI9341_T4::ILI9341Driver tft(PIN_CS, PIN_DC, PIN_SCK, PIN_MOSI, PIN_MISO, PIN_RESET, PIN_TOUCH_CS, PIN_TOUCH_IRQ);
// screen driver object
#ifdef USE_ST7789
ST7789_t3 tft = ST7789_t3(TFT_CS, TFT_DC, TFT_RST);
#elif defined(USE_ILI9488)
ILI9488_t3 tft = ILI9488_t3(TFT_CS, TFT_DC, TFT_RST);
#else
ILI9341_t3n tft = ILI9341_t3n(TFT_CS, TFT_DC, TFT_RST);
#endif

#define COLOR_BLUE 0x001F  /*   0,   0, 255 */
#define COLOR_GREEN 0x07E0 /*   0, 255,   0 */
#define COLOR_RED 0xF800   /* 255,   0,   0 */

// 2 x 10K diff buffers (used by tft) for differential updates (in DMAMEM)
//DMAMEM ILI9341_T4::DiffBuffStatic<6000> diff1;
//DMAMEM ILI9341_T4::DiffBuffStatic<6000> diff2;

// screen dimension (landscape mode)
static const int SLX = TFT_WIDTH;   //320;
static const int SLY = TFT_HEIGHT;  //240;

static const int CENTER_SLX = TFT_WIDTH / 2;   //320;
static const int CENTER_SLY = TFT_HEIGHT / 2;  //240;


// main screen framebuffer (150K in DTCM for fastest access)
uint16_t fb[SLX * SLY];

// internal framebuffer (150K in DMAMEM) used by the ILI9431_T4 library for double buffering.
DMAMEM uint16_t ib[TFT_WIDTH * TFT_HEIGHT];

// image that encapsulates fb.
Image<RGB565> im(fb, SLX, SLY);




/* draw the clock with given angle, scale and opacity */
void drawClock(float angle, float scale = 1.0f, float opacity = 1.0f) {
#ifdef USE_RGB32_IMAGE
    im.blitScaledRotated(antique_clock, { 120, 120 }, { CENTER_SLX, CENTER_SLY }, scale, angle, opacity);
#else
  if (opacity == 1.0f) {  // the clock does not use trnasparency so we can use the faster method when opacity = 1.0f
    im.blitScaledRotated(antique_clock, { 120, 120 }, { CENTER_SLX, CENTER_SLY }, scale, angle);
  } else {
    im.blitScaledRotated(antique_clock, { 120, 120 }, { CENTER_SLX, CENTER_SLY }, scale, angle, opacity);
#endif  
}


/* draw the small hand with given angle, scale and opacity */
void drawSmallHand(float angle, float scale = 1.0f, float opacity = 1.0f) {
  // always use the method with blending, even when opacity = 1.0f because the hand has transparency
  im.blitScaledRotated(small_hand, { 22, 117 }, { CENTER_SLX, CENTER_SLY }, 0.55f * scale, angle, opacity);
}


/* draw long hand with given angle, scale and opacity */
void drawLongHand(float angle, float scale = 1.0f, float opacity = 1.0f) {
  // always use the method with blending, even when opacity = 1.0f because the hand has transparency
  im.blitScaledRotated(long_hand, { 14, 197 }, { CENTER_SLX, CENTER_SLY }, 0.5f * scale, angle, opacity);
}




void setup() {
  Serial.begin(9600);

  // output debug infos to serial port.
  //  tft.output(&Serial);

// initialize the ILI9341 screen
#ifdef USE_ST7789
  tft.init(240, 320);  // Init ST7789 TFT_HEIGHTxTFT_WIDTH
#elif defined(USE_ILI9488)
  tft.begin(SPI_SPEED);
#else
  tft.begin(SPI_SPEED);
#endif

  tft.fillScreen(COLOR_RED);
  delay(500);
  tft.fillScreen(COLOR_GREEN);
  delay(500);
  tft.fillScreen(COLOR_BLUE);
  delay(500);
  // ok. turn on backlight
  pinMode(PIN_BACKLIGHT, OUTPUT);
  digitalWrite(PIN_BACKLIGHT, HIGH);

  // setup the screen driver
  tft.setRotation(3);        // portrait mode
  tft.setFrameBuffer(ib);    // double buffering set before turning on FB as to not have use malloc it
  tft.useFrameBuffer(true);  // turn on frame buffer
  //  tft.setDiffBuffers(&diff1, &diff2);  // 2 diff buffers
  //  tft.setDiffGap(4);                   // small gap
  //  tft.setRefreshRate(140);             // refresh at 60hz
  //  tft.setVSyncSpacing(2);
}



void update_display(bool force_whole_display) {
#if 1
  if (force_whole_display) {
    tft.setClipRect();
    tft.writeRect(IMAGE_CENTER, IMAGE_CENTER, SLX, SLY, fb);
    tft.updateScreen();
    return;
  }

// lets try writeRect
#if 1
  // look for bounding range of changes.
  uint16_t first_changed_x = tft.width();
  uint16_t last_changed_x = 0;
  uint16_t first_changed_y = tft.height();
  uint16_t last_changed_y = 0;

  for (uint16_t y = 0; y < tft.height(); y++) {
    uint16_t* ptftfb_y = &ib[y * tft.width()];
    uint16_t* pimage_newfb_y = &fb[y * tft.width()];
    uint16_t x;
    for (x = 0; x < tft.width(); x++) {
      // find first if any change in line.
      if (ptftfb_y[x] != pimage_newfb_y[x]) {
        if (first_changed_y == tft.height()) first_changed_y = y;
        last_changed_y = y;
        if (x < first_changed_x) first_changed_x = x;
        break;
      }
    }
    // find last change in line
    if (x < tft.width()) {
      for (x = tft.width() - 1; x > last_changed_x; x--) {
        if (ptftfb_y[x] != pimage_newfb_y[x]) {
          last_changed_x = x;  //
          break;
        }
      }
    }
  }
  // now have the bounds
  static uint8_t debug_count = 50;
  if (debug_count) {
    Serial.printf("(%u, %u), (%u, %u)\n", first_changed_x, last_changed_x, first_changed_y, last_changed_y);
    debug_count--;
  }
#endif
  tft.updateChangedAreasOnly(true);  // really only need to call this once.
  tft.writeRect(IMAGE_CENTER, IMAGE_CENTER, SLX, SLY, fb);
  tft.updateScreen();  // simply update.
#else
  if (!force_whole_display) {
    // look for bounding range of changes.
    uint16_t first_changed_x = tft.width();
    uint16_t last_changed_x = 0;
    uint16_t first_changed_y = tft.height();
    uint16_t last_changed_y = 0;

    for (uint16_t y = 0; y < tft.height(); y++) {
      uint16_t* ptftfb_y = &ib[y * tft.width()];
      uint16_t* pimage_newfb_y = &fb[y * tft.width()];
      uint16_t x;
      for (x = 0; x < tft.width(); x++) {
        // find first if any change in line.
        if (ptftfb_y[x] != pimage_newfb_y[x]) {
          if (first_changed_y == tft.height()) first_changed_y = y;
          last_changed_y = y;
          if (x < first_changed_x) first_changed_x = x;
          break;
        }
      }
      // find last change in line
      if (x < tft.width()) {
        for (x = tft.width() - 1; x > last_changed_x; x--) {
          if (ptftfb_y[x] != pimage_newfb_y[x]) {
            last_changed_x = x;  //
            break;
          }
        }
      }
    }
    // now have the bounds
    tft.setClipRect(first_changed_x, first_changed_y, (last_changed_x - first_changed_x) + 1, (last_changed_y - first_changed_y) + 1);
    static uint8_t debug_count = 20;
    if (debug_count) {
      Serial.printf("(%u, %u), (%u, %u)\n", first_changed_x, last_changed_x, first_changed_y, last_changed_y);
      debug_count--;
    }

  } else {
    tft.setClipRect();
  }
  memcpy(ib, fb, sizeof(fb));
  tft.updateScreen();  // simply update.
#endif
}


// used to keep track of the elapsed time
elapsedMillis em;

uint8_t loop_count = 0;

void loop() {
  int t;

  loop_count = (loop_count == 3)? 0: loop_count + 1;
#ifdef USE_RGB32_IMAGE
  const RGB32 background_color = RGB565_Black;
  switch (loop_count) {
    case 1: background_color = RGB32_Green; break;
    case 2: background_color = RGB32_Orange; break;
    case 3: background_color = RGB32_Silver; break;
  }
#else
  const RGB565 background_color = RGB565_Black;
  switch (loop_count) {
    case 1: background_color = RGB565_Green; break;
    case 2: background_color = RGB565_Orange; break;
    case 3: background_color = RGB565_Silver; break;
  }
#endif    
  Serial.println("----------------------------------------------");
  // PART 0 : title
  em = 0;
  while (em < 3000) {
    int y1 = (em < 1000) ? (130 * em) / 1000 - 50 : 80;
    int y2 = (em < 1000) ? 240 - (110 * em) / 1000 : 130;
    im.fillScreen(background_color);
    im.drawText("TGX library",iVec2{110,y1}, font_tgx_OpenSans_Bold_18, RGB565_Red);
    im.drawText("Crazy clock demo",iVec2{35,y2}, font_tgx_OpenSans_Bold_28, RGB565_White);
    update_display(false);
    yield();  // to keep the board responsive
  }

  // PART 1 : the clock appears.
  em = 0;
  while ((t = em) < 1000) {
    float sc = 0.1f + 0.9f * t / 1000.0f;
    im.fillScreen(background_color);
    im.drawText("TGX library",iVec2{110,80}, font_tgx_OpenSans_Bold_18, RGB565_Red, 1.0f - t/1000.0f);
    im.drawText("Crazy clock demo",iVec2{35,130}, font_tgx_OpenSans_Bold_28, RGB565_White, 1.0f - t/1000.0f);
    drawClock(200 - t / 5, sc);
    drawSmallHand(t / 10, sc);
    drawLongHand(t / 2, sc);
    update_display(false);
    yield();  // to keep the board responsive
  }

  // PART 2: rotation
  em = 0;
  while ((t = em) < 10000) {
    im.fillScreen(background_color);
    drawClock(0, 1.0f);
    drawSmallHand(100 + 360 * sin(t / 1500.0f), 1.0f);
    drawLongHand(500 * cos(t / 5000.0f), 1.0f);
    update_display(false);
    yield();  // to keep the board responsive
  }


#if defined(USE_ILI9488)
  em = 0;
  float scale = 320.0 / 240.0;
  while ((t = em) < 10000) {
    im.fillScreen(background_color);
    drawClock(0, scale);
    drawSmallHand(100 + 360 * sin(t / 1500.0f), scale);
    drawLongHand(500 * cos(t / 5000.0f), scale);
    update_display(false);
    yield();  // to keep the board responsive
  }
#endif

  // PART 3: changing hands sizes
  while ((t = em) < 20000) {
    im.fillScreen(background_color);
    drawClock(0, 1.0f);
    drawSmallHand(100 + 360 * sin(t / 1500.0f), 0.292893 + abs(sin(0.7853981 + (t - 10000.0f) / 3000.0f)));
    drawLongHand(500 * cos(t / 5000.0f), 0.292893 + abs(cos(0.7853981 + (t - 10000.0f) / 3000.0f)));
    update_display(false);
    yield();  // to keep the board responsive
  }

  // PART 4: rotating the whole clock
  while ((t = em) < 35000) {
    im.fillScreen(background_color);
    drawClock(150.0f * sin((t - 20000.0f) / 2000.0f), 1.0f);
    drawSmallHand(100 + 360 * sin(t / 1500.0f), 0.292893 + abs(sin(0.7853981 + (t - 10000.0f) / 3000.0f)));
    drawLongHand(500 * cos(t / 5000.0f), 0.292893 + abs(cos(0.7853981 + (t - 10000.0f) / 3000.0f)));
    update_display(false);
    yield();  // to keep the board responsive
  }

  // PART 5: fading
  while ((t = em) < 40000) {
    im.fillScreen(background_color);
    drawClock(150.0f * sin((t - 20000.0f) / 2000.0f), 1.0f - (t - 35000.0f) / 5000.0f, 1.0f - (t - 35000.0f) / 5000.0f);
    drawSmallHand(100 + 360 * sin(t / 1500.0f), 0.292893 + abs(sin(0.7853981 + (t - 10000.0f) / 3000.0f)), 1.0f - (t - 35000.0f) / 5000.0f);
    drawLongHand(500 * cos(t / 5000.0f), 0.292893 + abs(cos(0.7853981 + (t - 10000.0f) / 3000.0f)), 1.0f - (t - 35000.0f) / 5000.0f);
    update_display(false);
    yield();
  }

  // wait a bit before starting over.
  delay(2000);
}


/** end of file */
